/***************************************************************
 ************************* A TPP Project ***********************
 ******************** Visit http://tpp.dhs.org *****************
 ***************************************************************
aobject.cpp

Author(s): Paul Fleischer (polle@denmark.crosswinds.net)
Created: 22. July 1999

Copyright(c) 1999 The AKOS team.
All Rights Reserved.
****************************************************************/

#include "aobject.h"

/************************************************************************
 ******************************* aobject ********************************
 ************************************************************************/

//************************* Constructors ***************************


//The name-only constructor is for creating top-level objects, since it has
//a NULL parent.
aobject::aobject(char *newName)
{
  name = new dstring(newName);
  noOfChilds = 0;
  children = 0;
  parent = 0;
  type = 0;
  data = 0;
  spawned = false;
}

//The next constructor, creates an empty child-object, ie. it's parent isn't
//NULL 
aobject::aobject(char *newName, aobject *newParent)
{
  name = new dstring(newName);
  noOfChilds = 0;
  children = 0;
  spawned = false;

  type = 0;
  data = 0;

  parent = newParent;
  parent->atachChild(this);
}

//Creates a child-object, copying it's own childs from a template object.
//The new object becomes a child-object of the template object's parent,
//and the new object's parent is the same as the template object's parent
//Remeber that the new object is a COPY of the emplate, and not just
//an instance. All sub-objects are COPIES too.
//The result is a complete copy of the template object, the only difference is
//the name.
aobject::aobject(aobject *temp, char *newName)
{
  spawned = false;

  temp->dublicate(this);

  parent = temp->parent;

  //  parent->atachChild(this);

  //dublicate() sets the name to the same as the template object, but this
  //is wrong, we want to specify our own name
  type = temp;
  name = new dstring(newName);
}

//************************** Destructor *****************************
aobject::~aobject()
{
  //If this object has a parent, detach from it
  if(parent != NULL)
    parent->detachChild(this);

  if(data != NULL)
    data->deleteRef();

  //If this object has children, destroy them too.
  //Unfortunately this gives problem with win32, so maybe it's wrong, so
  //i'll keep it disabled until I am sure.
 /*  if(noOfChilds > 0)
    {
      for(unsigned int i=0;i<noOfChilds;i++)
	{
	  delete children[i];
	}
    }
  */
}

//************************ Public function **************************

//Returns a pointer to the name variable
dstring*
aobject::getName()
{
  return name;
}

//Returns a pointer to the child, specified by number(1 is the first)
aobject*
aobject::getChild(unsigned int no)
{
  if(no > noOfChilds)
    return 0;
  else
    return children[no-1];
}

//Returns a pointer to the child, specified by name.
aobject*
aobject::getChild(char *childName)
{
  //Search the chain for the child with the specified name...
  for(unsigned int i=0;i<noOfChilds;i++)
    {
      //...and return a pointer to the child, if found...
      if(strcmp(childName, children[i]->getName()->tochar()) == 0)
	return children[i];
    }
  //...else return a NULL pointer
  return 0;
}

//The spawnChild() functions, create child-objects without using variables
//This first spawnChild() function, creates a child-function to the current
//object, ie. the child's parent will be 'this'.
aobject*
aobject::spawnChild(char *childName)
{
  aobject *tmpChild;
  tmpChild = new aobject(childName, this);
  //  atachChild(tmpChild);
  tmpChild->spawned = true;
  //  tmpChild->data = "";
  return tmpChild;
}


//This spawnChild() function, is a bit more interesting.
//It creates a new object from a template(look above under constructors for
//explaination). The parent of the new child-object will be this.
aobject*
aobject::spawnChild(aobject *temp, char *childName)
{
  aobject *tmpChild;

  tmpChild = new aobject(temp, childName);
  tmpChild->parent = this;
  tmpChild->spawned = true;
  atachChild(tmpChild);
  return tmpChild;
}

//Deletes a child created by one of the spawnChild() commands.
bool
aobject::deleteChild(char *childName)
{
  //Get a pointer to the child to delete...
  aobject *tmpChild = getChild(childName);

  //If this pointer is NULL, the child does not exist, therefor return false...
  if(tmpChild == NULL)
    return false;

  //...If the child does exist, check if it is created by a spawnChild() 
  //function
  if(tmpChild->spawned == true)
    { 
      delete tmpChild;
      return true;
    }
  else
    return false;
}

//Return a pointer to the data
dstring*
aobject::getData()
{
  if(data != NULL)
    return data->getData();
  else
    return 0;
}

avalue*
aobject::getDataPointer()
{
  return data;
}

//Set the data to a dstring
void
aobject::setData(dstring *newValue)
{
  avalue *tmpdata;
  if(data == NULL)
    data = new avalue();

  tmpdata = data->setData(newValue->tochar());
  if(tmpdata != data)
    {
      delete data;
      data = tmpdata;
    }
  
}

//Set the data to a char*
void
aobject::setData(char *newValue)
{
  avalue *tmpdata;
  if(data == NULL)
    data = new avalue();

  tmpdata = data->setData(newValue);

  data = tmpdata;
}

void
aobject::setData(avalue *newValue)
{
  if(data != NULL)
    data->deleteRef();
  data = newValue;
  data->newRef();
}

unsigned int
aobject::getNoOfChilds()
{
  return noOfChilds;
}


aobject*
aobject::getType(void)
{
  return type;
}

aobject*
aobject::getSubObject(char *subName)
{
  bool stopLoop = false;
  aobject *assignObj = this;
  
  for(int i=0; !stopLoop; i++)
    {
      dstring obj("");
      for(; (subName[i] != 0) && (subName[i] != '.');i++)
	obj += subName[i];
      
      assignObj = assignObj->getChild(obj.tochar());
      
      if(assignObj == NULL)
	{
	  stopLoop = true;
	}

      if(subName[i] == 0)
	stopLoop = true;
    }

  return assignObj;
}

/*
  lookupObject() searches for the object with the name given by objname.
  It starts looking amongst it's own children, if the object is not found,
  it ask it's parent to do a lookupObject().
  This goes on until a NULL parent is encountered, in which case NULL is
  returned.
  If the object name contains one or more ".", then only the part before "." is
  checked amongst it's own children, if found the part after the "." will be
  sent to the child's lookupObject function.
 */

aobject*
aobject::lookupObject(char *objname)
{
  if(getSubObject(objname) == NULL)
    {
      if(parent != NULL)
	{
	  return parent->lookupObject(objname);
	}
      else
	return NULL;
    }
  else
    return getSubObject(objname);
}

void
aobject::getFullName(dstring *retName)
{
  dstring fullName("");

  if(parent != NULL)
    {
      parent->getFullName(&fullName);
      *retName += fullName.tochar();
      *retName += ".";
      *retName += name->tochar();
    }
  else
    *retName = name->tochar();
}

aobject*
aobject::getParent(void)
{
  return parent;
}

//*************************** Private functions **************************

//Atach a child to the objects child list.
void
aobject::atachChild(aobject *child)
{
  noOfChilds++;
  
  //Resize the dynamic array
  children = (aobject**)realloc(children, sizeof(aobject)*noOfChilds);

  children[noOfChilds-1] = child;
}

//Detach the child from the objects child list
void
aobject::detachChild(aobject *child)
{
  unsigned int childNo;

  //Where in the chain is the child?
  for(childNo=0; (childNo<noOfChilds) && (children[childNo] != child);childNo++);

  //If it isn't there... return
  if(children[childNo] != child)
    return;

  //If the child is the last in the chain, we can just call realloc,
  //and allocate less memory. However, if the child is not at the end
  //we need to...
  if(childNo < noOfChilds-1)
    {
      //...move children after it(the deleted child) in the chain, one up.... 
      //not exactly well explained, I know;)
      childNo++;
      for(;childNo < noOfChilds; childNo++)
	{
	  children[childNo-1] = children[childNo];
	}
    }
 
  noOfChilds--;

  //The last step, re-allocate the variable children, so that the last entry
  //in the chain is removed.
  children = (aobject**)realloc(children, sizeof(aobject)*noOfChilds);
}

//Dublicate the object and all it's sub-objects to the object specified by *copt.
void
aobject::dublicate(aobject *copy)
{
  
  copy->data = data; 

  if(data != NULL) 
    data->newRef();

  copy->spawned = spawned; /*I'm not sure if this is necesarry, but we want
			     a exact copy, so we better do it ;)*/
  copy->name = new dstring(name->tochar()); /*Again the .tochar() part, due 
					     to error in the dstring class.*/
  copy->noOfChilds = noOfChilds;

  copy->type = type;

  //Allocate memory for the children array
  copy->children = (aobject**)malloc(sizeof(aobject)*noOfChilds);

  for(unsigned int i=0;i<noOfChilds;i++)
    {
      //Allocate space for the new sub-object
      //(I have tried to use 'new' but it doesn't work, this however does ;)
      copy->children[i] = (aobject*)malloc(sizeof(aobject));

      //Just make sure that the allocation went well.
      //And if it did, dublicate the sub-object.
      if(copy->children[i] != NULL)
	children[i]->dublicate(copy->children[i]);
    }
}

/***************************************************************************
 ********************************** avalue *********************************
 ***************************************************************************/

//****************************** CONSTRUCTORS ******************************

avalue::avalue(void)
{
  refCount = 1;
}

//**************************** PUBLIC FUNCTION *****************************
avalue*
avalue::setData(char *newData)
{
  if(refCount > 1)
    {
      avalue *tmp = new avalue();
      tmp->setData(newData);
      deleteRef();
      return tmp;
    }
  else
    {
      data = newData;
      return this;
    }
}

dstring*
avalue::getData(void)
{
  return &data;
}

void
avalue::newRef(void)
{
  refCount++;
}

void
avalue::deleteRef(void)
{
  refCount--;
  if(refCount == 0)
    {
      delete this;
    }
}
